//
//  AMRollOverButton.m
//  PlateControl
//
//  Created by Andreas on Sat Jan 24 2004.
//  Copyright (c) 2004 Andreas Mayer. All rights reserved.
//
//	2004-03-17	- register forNSViewFrameDidChangeNotification to adjust
//				  the tracking rect; thanks go to Ville Virkkunen for reporting it
//				  (doesn't work in IB though - any help appreciated)
//	2004-05-09	- overwrite removeFromSuperview and
//				  removeFromSuperviewWithoutNeedingDisplay
//				  to remove the tracking rectangle.
//				  Thanks to Jeremy Dronfield for reporting the issue.
//	2004-08-07	- added acceptsFirstMouse: to return YES so that the button
//				  will work when window is not key.
//				  Thanks to Jon Flowers for reporting the issue
//	2005-06-02	- added shading support
//	2005-06-24	- added -viewDidMoveToWindow method
//	2005-07-11	- added -initWithCoder: and -encodeWithCoder: to encode
//				  am_doesNotTrackMouseInside instance variable
//	2005-07-18	- register for NSWindowDidResizeNotification
//	2008-04-29	- added support for keyed coding
//				- added controlShadow
//				- send setNeedsDisplay messages in methods that change visual appearance
//				- returns YES for -isFlipped
//	2011-08-08	- appearance is now encapsulated inside AMRollOverBottonAppearance
//					  each subclass of AMRollOverButton has its own distinct appearance
//					  accessors are +appearance and +setAppearance:
//					- buttonType and bezelStyle are explicitely overwritten from NSButton
//					- bordered is also overwritten
//					- adding support for Cocoa Auto Layout (not finished)


#import "AMRollOverButton.h"

float am_insetCell = 0.0;

@interface AMRollOverButtonCell (AMRollOverButton)
- (void)finishInit;
@end


@interface AMRollOverButton ()
+ (void)_loadAppearanceData;
- (void)am_setDefaultValues;
- (void)am_resetTrackingRect;
@end


@implementation AMRollOverButton

static NSMutableDictionary *appearances = nil;

// ======================================================
#pragma mark - class methods
// ======================================================

+ (void)initialize
{
	if (self == [AMRollOverButton class]) {
		appearances = [[NSMutableDictionary alloc] initWithCapacity:1];
	}
	[self _loadAppearanceData];
	[self setCellClass:[AMRollOverButtonCell class]];
}

+ (void)_loadAppearanceData
{
	NSString *resourceName = NSStringFromClass(self);
	NSURL *appearanceDataURL = [[NSBundle mainBundle] URLForResource:resourceName withExtension:@"button"];
	if ([appearanceDataURL isFileURL]) {
		NSError *theError = nil;
		NSData *appearanceData = [NSData dataWithContentsOfURL:appearanceDataURL options:0 error:&theError];
		if (!appearanceData) {
			NSLog(@"Error loading appearance data for %@: %@", resourceName, [theError description]);
		} else {
			AMRollOverButtonAppearance *theAppearance = [AMRollOverButtonAppearance appearanceWithDataRepresentation:appearanceData];
			[appearances setObject:theAppearance forKey:resourceName];
		}
	}
}

+ (AMRollOverButtonAppearance *)appearance
{
	AMRollOverButtonAppearance *result = [appearances objectForKey:NSStringFromClass(self)];
	return result;
}

+ (void)setAppearance:(AMRollOverButtonAppearance *)newAppearance
{
	NSString *resourceName = NSStringFromClass(self);
	[appearances setObject:newAppearance forKey:resourceName];
}


// ======================================================
#pragma mark - initialization
// ======================================================

- (id)initWithFrame:(NSRect)frameRect andTitle:(NSString *)theTitle
{
	if ((self = [super initWithFrame:frameRect])) {
		[self am_setDefaultValues];
		[self setTitle:theTitle];
	}
	return self;
}

- (void)awakeFromNib
{
	[self am_setDefaultValues];
}

- (void)am_setDefaultValues
{
	if (![[self cell] isKindOfClass:[AMRollOverButtonCell class]]) {
		NSString *title = [self title];
		if (title == nil) title = @"";
//		NSLog(@"create new AMRollOverButtonCell");
		// keep old cell around to copy some values
		NSButtonCell *oldCell = [[self cell] retain];
		[self setCell:[[[AMRollOverButtonCell alloc] initTextCell:title] autorelease]];
		[[self cell] setControlSize:NSRegularControlSize];
		[self.cell setTag:oldCell.tag];
		[self.cell setTarget:oldCell.target];
		[self.cell setAction:oldCell.action];
		[self.cell setContinuous:oldCell.isContinuous];
		[self.cell setState:self.state];
		[oldCell release]; // no longer needed
	}
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(frameDidChange:) name:NSViewFrameDidChangeNotification object:self];
	if (![self font]) {
		[self setFont:[NSFont boldSystemFontOfSize:[NSFont systemFontSizeForControlSize:NSRegularControlSize]]];
	}
	[self setPostsFrameChangedNotifications:YES];
	//[self am_resetTrackingRect];
}

- (void)am_resetTrackingRect
{
	if (am_trackingRect) {
		[self removeTrackingRect:am_trackingRect];
	}
	NSRect trackingRect = [self frame];
	trackingRect.origin = NSZeroPoint;
	if (!am_doesNotTrackMouseInside) {
		am_trackingRect = [self addTrackingRect:trackingRect owner:self userData:nil assumeInside:NO];
	} else {
		am_trackingRect = 0;
	}
}

- (id)initWithCoder:(NSCoder *)coder
{
	if ((self = [super initWithCoder:coder])) {
		if ([coder allowsKeyedCoding] && [coder containsValueForKey:@"am_doesNotTrackMouseInside"]) {
			// Can decode keys in any order
			am_doesNotTrackMouseInside = [coder decodeBoolForKey:@"am_doesNotTrackMouseInside"];
		} else {
//			NSLog(@"AMRollOverButton -initWithCoder:  (non-keyed archive)");
			// Must decode keys in same order as encodeWithCoder:
			[coder decodeValueOfObjCType:@encode(BOOL) at:&am_doesNotTrackMouseInside];
		}
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
	[super encodeWithCoder:coder];
	if ([coder allowsKeyedCoding]) {
		[coder encodeBool:am_doesNotTrackMouseInside forKey:@"am_doesNotTrackMouseInside"];
	} else {
		[coder encodeValueOfObjCType:@encode(BOOL) at:&am_doesNotTrackMouseInside];
	}	
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:NSWindowDidResizeNotification object:nil];
	[self removeTrackingRect:am_trackingRect];
	[super dealloc];
}


// ======================================================
#pragma mark - accessor methods
// ======================================================

- (BOOL)tracksMouseInside
{
	return !am_doesNotTrackMouseInside;
}

- (void)setTracksMouseInside:(BOOL)flag
{
	if (flag != !am_doesNotTrackMouseInside) {
		am_doesNotTrackMouseInside = !flag;
		if (!am_doesNotTrackMouseInside) {
			[self am_resetTrackingRect];
		} else {
			[self removeTrackingRect:am_trackingRect];
			if ([[self cell] mouseOver]) {
				[[self cell] setMouseOver:NO];
			}
		}
	}
}


- (NSControlSize)controlSize
{
	return [[self cell] controlSize];
}

- (void)setControlSize:(NSControlSize)newControlSize
{
	[[self cell] setControlSize:newControlSize];
	[self setNeedsDisplay];
}

- (NSColor *)controlColor
{
	return [[self cell] controlColor];
}

- (void)setControlColor:(NSColor *)newControlColor
{
	[[self cell] setControlColor:newControlColor];
	[self setNeedsDisplay];
}

- (NSColor *)shadingColor
{
	return [[self cell] shadingColor];
}

- (void)setShadingColor:(NSColor *)newShadingColor
{
	[[self cell] setShadingColor:newShadingColor];
	[self setNeedsDisplay];
}


- (NSColor *)frameColor
{
	return [[self cell] frameColor];
}

- (void)setFrameColor:(NSColor *)newFrameColor
{
	[[self cell] setFrameColor:newFrameColor];
	[self setNeedsDisplay];
}

- (NSColor *)textColor
{
	return [[self cell] textColor];
}

- (void)setTextColor:(NSColor *)newTextColor
{
	[[self cell] setTextColor:newTextColor];
	[self setNeedsDisplay];
}

- (NSColor *)arrowColor
{
	return [[self cell] arrowColor];
}

- (void)setArrowColor:(NSColor *)newArrowColor
{
	[[self cell] setArrowColor:newArrowColor];
	[self setNeedsDisplay];
}

- (NSColor *)mouseoverControlColor
{
	return [[self cell] mouseoverControlColor];
}

- (void)setMouseoverControlColor:(NSColor *)newMouseoverControlColor
{
	[[self cell] setMouseoverControlColor:newMouseoverControlColor];
	[self setNeedsDisplay];
}

- (NSColor *)mouseoverShadingColor
{
	return [[self cell] mouseoverShadingColor];
}

- (void)setMouseoverShadingColor:(NSColor *)newMouseoverShadingColor
{
	[[self cell] setMouseoverShadingColor:newMouseoverShadingColor];
	[self setNeedsDisplay];
}

- (NSColor *)mouseoverFrameColor
{
	return [[self cell] mouseoverFrameColor];
}

- (void)setMouseoverFrameColor:(NSColor *)newMouseoverFrameColor
{
	[[self cell] setMouseoverFrameColor:newMouseoverFrameColor];
	[self setNeedsDisplay];
}

- (NSColor *)mouseoverTextColor
{
	return [[self cell] mouseoverTextColor];
}

- (void)setMouseoverTextColor:(NSColor *)newMouseoverTextColor
{
	[[self cell] setMouseoverTextColor:newMouseoverTextColor];
	[self setNeedsDisplay];
}

- (NSColor *)mouseoverArrowColor
{
	return [[self cell] mouseoverArrowColor];
}

- (void)setMouseoverArrowColor:(NSColor *)newMouseoverArrowColor
{
	[[self cell] setMouseoverArrowColor:newMouseoverArrowColor];
	[self setNeedsDisplay];
}

- (NSColor *)highlightedControlColor
{
	return [[self cell] highlightedControlColor];
}

- (void)setHighlightedControlColor:(NSColor *)newHighlightedControlColor
{
	[[self cell] setHighlightedControlColor:newHighlightedControlColor];
	[self setNeedsDisplay];
}

- (NSColor *)highlightedShadingColor
{
	return [[self cell] highlightedShadingColor];
}

- (void)setHighlightedShadingColor:(NSColor *)newHighlightedShadingColor
{
	[[self cell] setHighlightedShadingColor:newHighlightedShadingColor];
	[self setNeedsDisplay];
}

- (NSColor *)highlightedFrameColor
{
	return [[self cell] highlightedFrameColor];
}

- (void)setHighlightedFrameColor:(NSColor *)newHighlightedFrameColor
{
	[[self cell] setHighlightedFrameColor:newHighlightedFrameColor];
	[self setNeedsDisplay];
}

- (NSColor *)highlightedTextColor
{
	return [[self cell] highlightedTextColor];
}

- (void)setHighlightedTextColor:(NSColor *)newHighlightedTextColor
{
	[[self cell] setHighlightedTextColor:newHighlightedTextColor];
	[self setNeedsDisplay];
}

- (NSColor *)highlightedArrowColor
{
	return [[self cell] highlightedArrowColor];
}

- (void)setHighlightedArrowColor:(NSColor *)newHighlightedArrowColor
{
	[[self cell] setHighlightedArrowColor:newHighlightedArrowColor];
	[self setNeedsDisplay];
}

- (NSShadow *)controlShadow
{
	return [[self cell] controlShadow];
}

- (void)setControlShadow:(NSShadow *)newControlShadow
{
	[[self cell] setControlShadow:newControlShadow];
	[self setNeedsDisplay];
}

- (NSShadow *)textShadow
{
	return [[self cell] textShadow];
}

- (void)setTextShadow:(NSShadow *)newTextShadow
{
	[[self cell] setTextShadow:newTextShadow];
	[self setNeedsDisplay];
}

- (NSShadow *)mouseoverTextShadow
{
	return [[self cell] mouseoverTextShadow];
}

- (void)setMouseoverTextShadow:(NSShadow *)newMouseoverTextShadow
{
	[[self cell] setMouseoverTextShadow:newMouseoverTextShadow];
	[self setNeedsDisplay];
}

- (NSShadow *)highlightedTextShadow
{
	return [[self cell] highlightedTextShadow];
}

- (void)setHighlightedTextShadow:(NSShadow *)newHighlightedTextShadow
{
	[[self cell] setHighlightedTextShadow:newHighlightedTextShadow];
	[self setNeedsDisplay];
}

- (NSShadow *)highlightedControlShadow
{
	return [[self cell] highlightedControlShadow];
}

- (void)setHighlightedControlShadow:(NSShadow *)newHighlightedControlShadow
{
	[[self cell] setHighlightedControlShadow:newHighlightedControlShadow];
	[self setNeedsDisplay];
}

- (double)popUpMenuDelay
{
    return [[self cell] popUpMenuDelay];
}

- (void)setPopUpMenuDelay:(double)newPopUpMenuDelay
{
	[[self cell] setPopUpMenuDelay:newPopUpMenuDelay];
}

- (NSButtonType)buttonType
{
	return am_buttonType;
}

- (void)setButtonType:(NSButtonType)newButtonType
{
	am_buttonType = newButtonType;
	[super setButtonType:newButtonType];
	[self setNeedsDisplay];
}

- (NSBezelStyle)bezelStyle
{
	if ([[self cell] isKindOfClass:[AMRollOverButtonCell class]]) {
		return [[self cell] amBezelStyle];
	} else {
		return [super bezelStyle];
	}
}

- (void)setBezelStyle:(NSBezelStyle)newBezelStyle
{
	if ([[self cell] isKindOfClass:[AMRollOverButtonCell class]]) {
		[[self cell] setAmBezelStyle:newBezelStyle];
	} else {
		[super setBezelStyle:newBezelStyle];
	}
}

- (AMRollOverButtonShadingMode)shadingMode
{
	return [[self cell] shadingMode];
}

- (void)setShadingMode:(AMRollOverButtonShadingMode)newShadingMode
{
	[[self cell] setShadingMode:newShadingMode];
	[self setNeedsDisplay];
}

- (BOOL)isBordered
{
	return [[self cell] isBordered];
}

- (void)setBordered:(BOOL)flag
{
	[[self cell] setBordered:flag];
}

// ======================================================
#pragma mark - NSButton and NSView methods
// ======================================================

- (BOOL)isOpaque
{
	return NO;
}

- (BOOL)isFlipped
{
	return YES;
}

- (void)mouseEntered:(NSEvent *)theEvent
{
	if ([self isEnabled]) {
		[[self cell] setMouseOver:YES];
		[self setNeedsDisplay:YES];
	}
}

- (void)mouseExited:(NSEvent *)theEvent
{
	if ([self isEnabled]) {
		[[self cell] setMouseOver:NO];
		[self setNeedsDisplay:YES];
	}
}

- (void)drawRect:(NSRect)aRect
{
	NSRect cellRect = [self frame];
	cellRect.origin = NSZeroPoint;
	cellRect = NSInsetRect(cellRect, am_insetCell, am_insetCell);
	[[self cell] drawWithFrame:cellRect inView:self];
}

- (void)sizeToFit
{
	NSRect newFrame = [self frame];
	newFrame.size = [self.cell cellSize];
	newFrame.size.width += 2.0*am_insetCell;
	newFrame.size.height += 2.0*am_insetCell;
	[self setFrame:newFrame];
}

- (void)frameDidChange:(NSNotification *)aNotification
{
//	NSLog(@"frameDidChange: %f %f %f %f", self.frame.origin.x, self.frame.origin.y, self.frame.size.width, self.frame.size.height);
	[self am_resetTrackingRect];
}

- (void)removeFromSuperview
{
	if (am_trackingRect) {
		[self removeTrackingRect:am_trackingRect];
	}
	[super removeFromSuperview];
}

- (void)removeFromSuperviewWithoutNeedingDisplay
{
	if (am_trackingRect) {
		[self removeTrackingRect:am_trackingRect];
	}
	[super removeFromSuperviewWithoutNeedingDisplay];
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
	return YES;
}

- (void)viewDidMoveToWindow
{
	if ([self window]) {
		[self am_resetTrackingRect];
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(frameDidChange:) name:NSWindowDidResizeNotification object:[self window]];
	} else {
		[self removeTrackingRect:am_trackingRect];
		[[NSNotificationCenter defaultCenter] removeObserver:self name:NSWindowDidResizeNotification object:nil];
	}
}

- (void)setEnabled:(BOOL)flag
{
	if (!flag) {
		[[self cell] setMouseOver:NO];
		[self setNeedsDisplay:YES];
	}
	[super setEnabled:flag];
}

- (BOOL)canBecomeKeyView
{
	return NO;
}

// ======================================================
#pragma mark - auto layout methods
// ======================================================

- (NSSize)intrinsicContentSize
{
	NSSize result = [super intrinsicContentSize];
	result.width = [(AMRollOverButtonCell *)[self cell] intrinsicContentSize].width;
	return result;
}

/* doesn't give correct result yet
- (CGFloat)baselineOffsetFromBottom
{
	CGFloat result = 0.0;
	return result;
}
 */


@end
