//
//	AMRollOverButtonCell.m
//	PlateControl
//
//	Created by Andreas on Sat Jan 24 2004.
//	Copyright (c) 2004 Andreas Mayer. All rights reserved.
//
//	2004-02-01	- cleaned up layout code
//					- open menu instantly when user clicks on arrow
//	2004-03-15	- Don McConnel sent me a better formula for calculating
//					  the textInset which doesn't need trigonometric functions,
//					  so we don't need to include the libmx.dylib - thanks Don!
//	2004-07-08	- set textShadow to an empty NSShadow object instead
//					  of nil
//	2005-05-30	- hack to account for System Font xHeight change in 10.4:
//					  hard coded system font sizes
//	2005-06-02	- added shading support
//	2005-06-03	- modified calculation for text position to center on capHeight;
//					  (we keep the previous xHeight hack for calculating the horizontal inset)
//	2005-06-24	- honors the font set through -setFont:; will still adjust the font size
//					  when changing the control size
//	2005-07-11	- modified text positioning code again
//					- added second bezel style:
//					  NSShadowlessSquareBezelStyle will result in a button the shape of a
//					  rounded rectangle instead of the usual pill shape
//	2005-07-14	- added disabled look
//	2005-11-15	- replaced nested function showMenu by method
//	2006-01-07	- send action method from control instead of from cell
//					- support NSOnOffButton button type
//	2006-01-12	- fixed menu arrow position
//	2006-04-17	- -drawWithFrame:inView: honors button state (on/off)
//	2006-11-17	- -widthForFrame: should return more sensible values now
//	2007-04-21	- -widthForFrame: rounds value to nearest integer (thanks to Steve Israelson)
//	2008-04-29	- added controlShadow
//					- renamed -showArrow to -alwaysShowMenuIndicator
//					- added -amRollOverButtonDefaultHighlightedControlShadow
//					- added -copyWithZone: implementation
//					- added support for keyed coding
//					- accessor methods will now retain-autorelease
//					- setters for shadows will now copy new values
//					- use accessor methods in code instead of accessing variables directly
//					- implemented -setFont: to trigger re-layout
//					- fixed left/right text margins for rectangular button layout
//					- use -showsStateBy and highlightsBy to determine button behaviour
//					  should now behave like any other NSButton, no matter what type you choose
//					- frame is now drawn *on top* of the buttons background;
//					  you may need to change your frame's color to adjust for it
//	2011-08-08	- moved appearance data to AMRollOverButtonAppearance
//					- removed buttonType property, since that is part of the behaviour which should not be
//					  declared by the cell
//					- removed support for non-keyed coders
//					- re-added support for On-Off-Buttons (got lost somewhere ...)
//					- removed dependencies on some NSBotton(Cell) internals


//#import <Carbon/Carbon.h>
#import "AMRollOverButtonCell.h"
#import "AMRollOverButtonAppearance.h"
#import "AMRollOverButton.h"
#import "NSBezierPath_AMAdditons.h"
#import "NSBezierPath_AMShading.h"
#import "NSColor_AMAdditions.h"
#import "NSFont_AMFixes.h"
#import </usr/include/math.h>

static float am_backgroundInset = 1.5;
static float am_textGap = 1.5;
//static float am_textInsetFactor = 1.3;
static float am_roundedRectCornerRadius = 4.3;
static float am_bezierPathFlatness = 0.2;

// ======================================================
#pragma mark - Default Shadow Values NSShadow category
// ======================================================

@interface NSShadow (AMRollOverButton)
+ (NSShadow *)amRollOverButtonDefaultControlShadow;
+ (NSShadow *)amRollOverButtonDefaultHighlightedControlShadow;
+ (NSShadow *)amRollOverButtonDefaultTextShadow;
@end

@implementation NSShadow (AMRollOverButton)

+ (NSShadow *)amRollOverButtonDefaultControlShadow
{
	NSShadow *result = [[[NSShadow alloc] init] autorelease];
	[result setShadowOffset:NSMakeSize(0.0, -1.0)];
	[result setShadowBlurRadius:1.0];
	[result setShadowColor:[NSColor controlLightHighlightColor]];
	return result;
}

+ (NSShadow *)amRollOverButtonDefaultHighlightedControlShadow
{
	NSShadow *result = [[[NSShadow alloc] init] autorelease];
	[result setShadowOffset:NSMakeSize(0.0, 1.0)];
	[result setShadowBlurRadius:1.0];
	[result setShadowColor:[NSColor controlDarkShadowColor]];
	return result;
}

+ (NSShadow *)amRollOverButtonDefaultTextShadow
{
	NSShadow *result = [[[NSShadow alloc] init] autorelease];
	[result setShadowOffset:NSMakeSize(0.0, -1.0)];
	[result setShadowBlurRadius:1.0];
	[result setShadowColor:[NSColor colorWithCalibratedWhite:0.3 alpha:1.0]];
	return result;
}

@end

// ======================================================
#pragma mark - Rounded Rect NSBezierPath category
// ======================================================

@interface NSBezierPath (AMRollOverButton)
+ (NSBezierPath *)bezierPathForButtonWithRoundedRect:(NSRect)rect;
@end

@implementation NSBezierPath (AMRollOverButton)

+ (NSBezierPath *)bezierPathForButtonWithRoundedRect:(NSRect)rect
{
	return [NSBezierPath bezierPathWithRoundedRect:rect cornerRadius:am_roundedRectCornerRadius];
}

@end

// ======================================================
#pragma mark - Class extension
// ======================================================

@interface AMRollOverButtonCell ()
- (NSSize)lastFrameSize;
- (void)setLastFrameSize:(NSSize)newLastFrameSize;
- (void)finishInit;
- (CGFloat)textInsetForHeight:(CGFloat)height;
@end


// ======================================================
#pragma mark - Implementation
// ======================================================

@implementation AMRollOverButtonCell

// ======================================================
#pragma mark - Initialization
// ======================================================

- (id)init
{
	if ((self = [super init])) {
		[self finishInit];
	}
	return self;
}

- (id)initTextCell:(NSString *)aString
{
	if ((self = [super initTextCell:aString])) {
		[self finishInit];
	}
	return self;
}

- (id)copyWithZone:(NSZone*)aZone
{
	
	/*
	 This call is provided mostly for completeness-sake.  It used to be
	 the case where it would called in the context of older IB palettes.
	 
	 But, it now appears to be the case where the NSCoding protocol APIs
	 are used by IB to duplicate, copy-paste, etc. instances of this class.
	 
	 AM: Turns out it is still used - at least when editing cells.
	 */
	
	AMRollOverButtonCell *result = [[AMRollOverButtonCell allocWithZone:aZone] init];
	if (_appearance) {
		[result setControlColor:[[[self controlColor] copyWithZone:aZone] autorelease]];
		[result setFrameColor:[[[self frameColor] copyWithZone:aZone] autorelease]];
		[result setTextColor:[[[self textColor] copyWithZone:aZone] autorelease]];
		[result setArrowColor:[[[self arrowColor] copyWithZone:aZone] autorelease]];
		[result setShadingColor:[[[self shadingColor] copyWithZone:aZone] autorelease]];
		[result setMouseoverControlColor:[[[self mouseoverControlColor] copyWithZone:aZone] autorelease]];
		[result setMouseoverFrameColor:[[[self mouseoverFrameColor] copyWithZone:aZone] autorelease]];
		[result setMouseoverTextColor:[[[self mouseoverTextColor] copyWithZone:aZone] autorelease]];
		[result setMouseoverArrowColor:[[[self mouseoverArrowColor] copyWithZone:aZone] autorelease]];
		[result setMouseoverShadingColor:[[[self mouseoverShadingColor] copyWithZone:aZone] autorelease]];
		[result setHighlightedControlColor:[[[self highlightedControlColor] copyWithZone:aZone] autorelease]];
		[result setHighlightedFrameColor:[[[self highlightedFrameColor] copyWithZone:aZone] autorelease]];
		[result setHighlightedTextColor:[[[self highlightedTextColor] copyWithZone:aZone] autorelease]];
		[result setHighlightedArrowColor:[[[self highlightedArrowColor] copyWithZone:aZone] autorelease]];
		[result setHighlightedShadingColor:[[[self highlightedShadingColor] copyWithZone:aZone] autorelease]];
		[result setControlShadow:[[[self controlShadow] copyWithZone:aZone] autorelease]];
		[result setTextShadow:[[[self textShadow] copyWithZone:aZone] autorelease]];
		[result setMouseoverTextShadow:[[[self mouseoverTextShadow] copyWithZone:aZone] autorelease]];
		[result setHighlightedTextShadow:[[[self highlightedTextShadow] copyWithZone:aZone] autorelease]];
		[result setHighlightedControlShadow:[[[self highlightedControlShadow] copyWithZone:aZone] autorelease]];
	}	
	return result;
}

- (void)finishInit
{
	[self setAlignment:NSCenterTextAlignment];
}

- (id)initWithCoder:(NSCoder *)coder
{
	if ((self = [super initWithCoder:coder])) {
		if ([coder allowsKeyedCoding]) {
			// Can decode keys in any order
			if ([coder containsValueForKey:@"am_controlColor"]) {
				_appearance = [[AMRollOverButtonAppearance alloc] init];
				[self setControlColor:[coder decodeObjectForKey:@"am_controlColor"]];
				[self setFrameColor:[coder decodeObjectForKey:@"am_frameColor"]];
				[self setTextColor:[coder decodeObjectForKey:@"am_textColor"]];
				[self setArrowColor:[coder decodeObjectForKey:@"am_arrowColor"]];
				[self setMouseoverControlColor:[coder decodeObjectForKey:@"am_mouseoverControlColor"]];
				[self setMouseoverFrameColor:[coder decodeObjectForKey:@"am_mouseoverFrameColor"]];
				[self setMouseoverTextColor:[coder decodeObjectForKey:@"am_mouseoverTextColor"]];
				[self setMouseoverArrowColor:[coder decodeObjectForKey:@"am_mouseoverArrowColor"]];
				[self setHighlightedControlColor:[coder decodeObjectForKey:@"am_highlightedControlColor"]];
				[self setHighlightedFrameColor:[coder decodeObjectForKey:@"am_highlightedFrameColor"]];
				[self setHighlightedTextColor:[coder decodeObjectForKey:@"am_highlightedTextColor"]];
				[self setHighlightedArrowColor:[coder decodeObjectForKey:@"am_highlightedArrowColor"]];
				[self setControlShadow:[coder decodeObjectForKey:@"am_controlShadow"]];
				[self setTextShadow:[coder decodeObjectForKey:@"am_textShadow"]];
				[self setMouseoverTextShadow:[coder decodeObjectForKey:@"am_mouseoverTextShadow"]];
				[self setHighlightedTextShadow:[coder decodeObjectForKey:@"am_highlightedTextShadow"]];
				[self setHighlightedControlShadow:[coder decodeObjectForKey:@"am_highlightedControlShadow"]];
				[self setPopUpMenuDelay:[coder decodeDoubleForKey:@"am_popUpMenuDelay"]];
				[self setShadingColor:[coder decodeObjectForKey:@"am_shadingColor"]];
				[self setMouseoverShadingColor:[coder decodeObjectForKey:@"am_mouseoverShadingColor"]];
				[self setHighlightedShadingColor:[coder decodeObjectForKey:@"am_highlightedShadingColor"]];
				[self setShadingMode:[coder decodeIntegerForKey:@"am_shadingMode"]];
//				if ([coder containsValueForKey:@"am_buttonType"]) {
//					[self setButtonType:[coder decodeIntForKey:@"am_buttonType"]];
//				}
				[self setAmBezelStyle:[coder decodeIntegerForKey:@"am_bezelStyle"]];
				[self setAlwaysShowMenuIndicator:[coder decodeBoolForKey:@"am_alwaysShowMenuIndicator"]];
			}
		} else {
			NSLog(@"will only work with keyed coding");
		}	
		[self setAlignment:NSCenterTextAlignment];
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
	[super encodeWithCoder:coder];
	if ([coder allowsKeyedCoding]) {
		if (_appearance) {
			[coder encodeObject:[self controlColor] forKey:@"am_controlColor"];
			[coder encodeObject:[self frameColor] forKey:@"am_frameColor"];
			[coder encodeObject:[self textColor] forKey:@"am_textColor"];
			[coder encodeObject:[self arrowColor] forKey:@"am_arrowColor"];
			[coder encodeObject:[self mouseoverControlColor] forKey:@"am_mouseoverControlColor"];
			[coder encodeObject:[self mouseoverFrameColor] forKey:@"am_mouseoverFrameColor"];
			[coder encodeObject:[self mouseoverTextColor] forKey:@"am_mouseoverTextColor"];
			[coder encodeObject:[self mouseoverArrowColor] forKey:@"am_mouseoverArrowColor"];
			[coder encodeObject:[self highlightedControlColor] forKey:@"am_highlightedControlColor"];
			[coder encodeObject:[self highlightedFrameColor] forKey:@"am_highlightedFrameColor"];
			[coder encodeObject:[self highlightedTextColor] forKey:@"am_highlightedTextColor"];
			[coder encodeObject:[self highlightedArrowColor] forKey:@"am_highlightedArrowColor"];
			[coder encodeObject:[self controlShadow] forKey:@"am_controlShadow"];
			[coder encodeObject:[self textShadow] forKey:@"am_textShadow"];
			[coder encodeObject:[self mouseoverTextShadow] forKey:@"am_mouseoverTextShadow"];
			[coder encodeObject:[self highlightedTextShadow] forKey:@"am_highlightedTextShadow"];
			[coder encodeObject:[self highlightedControlShadow] forKey:@"am_highlightedControlShadow"];
			[coder encodeDouble:[self popUpMenuDelay] forKey:@"am_popUpMenuDelay"];
			[coder encodeObject:[self shadingColor] forKey:@"am_shadingColor"];
			[coder encodeObject:[self mouseoverShadingColor] forKey:@"am_mouseoverShadingColor"];
			[coder encodeObject:[self highlightedShadingColor] forKey:@"am_highlightedShadingColor"];
			[coder encodeInteger:[self shadingMode] forKey:@"am_shadingMode"];
			[coder encodeInteger:[self amBezelStyle] forKey:@"am_bezelStyle"];
			[coder encodeBool:[self alwaysShowMenuIndicator] forKey:@"am_alwaysShowMenuIndicator"];
		}
	} else {
		NSLog(@"will only work with keyed coding");
	}	
}

- (void)dealloc
{
	[_appearance release];
	[am_arrowPath release];
	[am_backgroundPath release];
	[am_highlightedBackgroundPath release];
	[super dealloc];
}


// ======================================================
#pragma mark - Accessor methods
// ======================================================

- (void)setMenu:(NSMenu *)menu
{
	[super setMenu:menu];
	am_lastFrameSize.width = 0.0;
}

- (AMRollOverButtonAppearance *)appearance
{
	AMRollOverButtonAppearance *result = _appearance;
	if (!result) {
		result = [[(AMRollOverButton *)[self controlView] class] appearance];
	}
	return result;
}

- (void)setAppearance:(AMRollOverButtonAppearance *)newAppearance
{
	AMRollOverButtonAppearance *oldAppearance = _appearance;
	_appearance = [newAppearance retain];
	[oldAppearance release];
}

- (NSColor *)controlColor
{
	return [self.appearance controlColor];
}

- (void)setControlColor:(NSColor *)newControlColor
{
	[self.appearance setControlColor:newControlColor];
}

- (NSColor *)frameColor
{
	return [self.appearance frameColor];
}

- (void)setFrameColor:(NSColor *)newFrameColor
{
	[self.appearance setFrameColor:newFrameColor];
}

- (NSColor *)textColor
{
	return [self.appearance textColor];
}

- (void)setTextColor:(NSColor *)newTextColor
{
	[self.appearance setTextColor:newTextColor];
}

- (NSColor *)arrowColor
{
	return [self.appearance arrowColor];
}

- (void)setArrowColor:(NSColor *)newArrowColor
{
	[self.appearance setArrowColor:newArrowColor];
}

- (NSColor *)shadingColor
{
	return [self.appearance shadingColor];
}

- (void)setShadingColor:(NSColor *)newShadingColor
{
	[self.appearance setShadingColor:newShadingColor];
}


- (NSColor *)mouseoverControlColor
{
	return [self.appearance mouseoverControlColor];
}

- (void)setMouseoverControlColor:(NSColor *)newMouseoverControlColor
{
	[self.appearance setMouseoverControlColor:newMouseoverControlColor];
}

- (NSColor *)mouseoverFrameColor
{
	return [self.appearance mouseoverFrameColor];
}

- (void)setMouseoverFrameColor:(NSColor *)newMouseoverFrameColor
{
	[self.appearance setMouseoverFrameColor:newMouseoverFrameColor];
}

- (NSColor *)mouseoverTextColor
{
	return [self.appearance mouseoverTextColor];
}

- (void)setMouseoverTextColor:(NSColor *)newMouseoverTextColor
{
	[self.appearance setMouseoverTextColor:newMouseoverTextColor];
}

- (NSColor *)mouseoverArrowColor
{
	return [self.appearance mouseoverArrowColor];
}

- (void)setMouseoverArrowColor:(NSColor *)newMouseoverArrowColor
{
	[self.appearance setMouseoverArrowColor:newMouseoverArrowColor];
}

- (NSColor *)mouseoverShadingColor
{
	return [self.appearance mouseoverShadingColor];
}

- (void)setMouseoverShadingColor:(NSColor *)newMouseoverShadingColor
{
	[self.appearance setMouseoverShadingColor:newMouseoverShadingColor];
}


- (NSColor *)highlightedControlColor
{
	return [self.appearance highlightedControlColor];
}

- (void)setHighlightedControlColor:(NSColor *)newHighlightedControlColor
{
	[self.appearance setHighlightedControlColor:newHighlightedControlColor];
}

- (NSColor *)highlightedFrameColor
{
	return [self.appearance highlightedFrameColor];
}

- (void)setHighlightedFrameColor:(NSColor *)newHighlightedFrameColor
{
	[self.appearance setHighlightedFrameColor:newHighlightedFrameColor];
}

- (NSColor *)highlightedTextColor
{
	return [self.appearance highlightedTextColor];
}

- (void)setHighlightedTextColor:(NSColor *)newHighlightedTextColor
{
	[self.appearance setHighlightedTextColor:newHighlightedTextColor];
}

- (NSColor *)highlightedArrowColor
{
	return [self.appearance highlightedArrowColor];
}

- (void)setHighlightedArrowColor:(NSColor *)newHighlightedArrowColor
{
	[self.appearance setHighlightedArrowColor:newHighlightedArrowColor];
}

- (NSColor *)highlightedShadingColor
{
	return [self.appearance highlightedShadingColor];
}

- (void)setHighlightedShadingColor:(NSColor *)newHighlightedShadingColor
{
	[self.appearance setHighlightedShadingColor:newHighlightedShadingColor];
}

- (NSShadow *)controlShadow;
{
	return [self.appearance controlShadow];
}

- (void)setControlShadow:(NSShadow *)newControlShadow;
{
	[self.appearance setControlShadow:newControlShadow];
}

- (NSShadow *)textShadow
{
	return [self.appearance textShadow];
}

- (void)setTextShadow:(NSShadow *)newTextShadow
{
	[self.appearance setTextShadow:newTextShadow];
}

- (NSShadow *)mouseoverTextShadow
{
	return [self.appearance mouseoverTextShadow];
}

- (void)setMouseoverTextShadow:(NSShadow *)newMouseoverTextShadow
{
	[self.appearance setMouseoverTextShadow:newMouseoverTextShadow];
}

- (NSShadow *)highlightedTextShadow
{
	return [self.appearance highlightedTextShadow];
}

- (void)setHighlightedTextShadow:(NSShadow *)newHighlightedTextShadow
{
	[self.appearance setHighlightedTextShadow:newHighlightedTextShadow];
}

- (NSShadow *)highlightedControlShadow
{
	return [self.appearance highlightedControlShadow];
}

- (void)setHighlightedControlShadow:(NSShadow *)newHighlightedControlShadow
{
	[self.appearance setHighlightedControlShadow:newHighlightedControlShadow];
}

- (double)popUpMenuDelay
{
	return [self.appearance popUpMenuDelay];
}

- (void)setPopUpMenuDelay:(double)newPopUpMenuDelay
{
	[self.appearance setPopUpMenuDelay:newPopUpMenuDelay];
}

- (AMRollOverButtonShadingMode)shadingMode
{
	return [self.appearance shadingMode];
}

- (void)setShadingMode:(AMRollOverButtonShadingMode)newShadingMode
{
	[self.appearance setShadingMode:newShadingMode];
}


- (NSBezierPath *)backgroundPath
{
	return [[am_backgroundPath retain] autorelease];
}

- (void)setBackgroundPath:(NSBezierPath *)newBackgroundPath
{
	id old = nil;

	if (newBackgroundPath != am_backgroundPath) {
		old = am_backgroundPath;
		am_backgroundPath = [newBackgroundPath retain];
		[old release];
	}
}

- (NSBezierPath *)highlightedBackgroundPath
{
	return [[am_highlightedBackgroundPath retain] autorelease];
}

- (void)setHighlightedBackgroundPath:(NSBezierPath *)newHighlightedBackgroundPath
{
	id old = nil;

	if (newHighlightedBackgroundPath != am_highlightedBackgroundPath) {
		old = am_highlightedBackgroundPath;
		am_highlightedBackgroundPath = [newHighlightedBackgroundPath retain];
		[old release];
	}
}

- (NSBezierPath *)arrowPath
{
	return [[am_arrowPath retain] autorelease];
}

- (void)setArrowPath:(NSBezierPath *)newArrowPath
{
	id old = nil;
	
	if (newArrowPath != am_arrowPath) {
		old = am_arrowPath;
		am_arrowPath = [newArrowPath retain];
		[old release];
	}
}

- (NSSize)lastFrameSize
{
	return am_lastFrameSize;
}

- (void)setLastFrameSize:(NSSize)newLastFrameSize
{
	am_lastFrameSize = newLastFrameSize;
}

- (BOOL)alwaysShowMenuIndicator
{
	return [self.appearance alwaysShowMenuIndicator];
}

- (void)setAlwaysShowMenuIndicator:(BOOL)newValue
{
	[self.appearance setAlwaysShowMenuIndicator:newValue];
}

- (BOOL)mouseOver
{
	return am_mouseOver;
}

- (void)setMouseOver:(BOOL)newMouseOver
{
	am_mouseOver = newMouseOver;
}

- (NSBezelStyle)amBezelStyle
{
	return [self.appearance bezelStyle];
}

- (void)setBezelStyle:(NSBezelStyle)newBezelStyle
{
	[self setAmBezelStyle:newBezelStyle];
}

- (void)setAmBezelStyle:(NSBezelStyle)newAmBezelStyle
{
	[self.appearance setBezelStyle:newAmBezelStyle];
	am_lastFrameSize.width = 0;
}

- (BOOL)isBordered
{
	return [self.appearance isBordered];
}

- (void)setBordered:(BOOL)flag
{
	[self.appearance setBordered:flag];
}

- (void)setFont:(NSFont *)newFont
{
	[super setFont:newFont];
	am_lastFrameSize.width = 0;
}

// ======================================================
#pragma mark - Layout
// ======================================================

- (void)invalidateLayout
{
	am_lastFrameSize.width = 0;
}

- (void)calculateLayoutForFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	// bezier path for plate background
	[self setLastFrameSize:cellFrame.size];
	NSRect innerRect = NSInsetRect(cellFrame, am_backgroundInset, am_backgroundInset);
	innerRect.size.height -= 1;
	if (![controlView isFlipped]) {
		innerRect.origin.y -= 1;
	}
	// text rect
	am_textRect = innerRect;
	NSFont *font = [self font];
	NSDictionary *stringAttributes = [NSDictionary dictionaryWithObjectsAndKeys:font, NSFontAttributeName, nil];
	NSAttributedString *string = [[[NSAttributedString alloc] initWithString:[self title] attributes:stringAttributes] autorelease];
	NSSize size = [string size];
	float radius = (am_lastFrameSize.height/2.0)-am_backgroundInset;
	// calculate minimum text inset
	float textInset = [self textInsetForHeight:am_lastFrameSize.height];
	am_textRect = NSInsetRect(am_textRect, textInset+am_textGap, 0); // 
	
	am_textRect.size.height = size.height;
	float capHeight = [font fixed_capHeight];
	float ascender = [font ascender];
	float descender = -[font descender];
	float yOrigin = innerRect.origin.y;
	if ([controlView isFlipped]) {
		// line gap below descender
		float offset = ((innerRect.size.height-am_textRect.size.height) / 2.0);
		offset += (ascender-capHeight)-((am_textRect.size.height-capHeight) / 2.0);
		yOrigin += ceilf(offset);
	} else {
		// line gap on top of ascender
		float offset = ceilf((innerRect.size.height-am_textRect.size.height) / 2.0);
		offset += (((am_textRect.size.height-capHeight) / 2.0)-descender);
		yOrigin += offset;
	}
	am_textRect.origin.y = yOrigin;
//	NSLog(@"text rect width: %f", am_textRect.size.width);

	// bezier path for button background
	innerRect.origin.x = 0;
	innerRect.origin.y = 0;

	if ([self amBezelStyle] == NSShadowlessSquareBezelStyle) {
		am_getBackgroundSelector = @selector(bezierPathForButtonWithRoundedRect:);
	} else {
		am_getBackgroundSelector = @selector(bezierPathWithPlateInRect:);
	}
		
	id returnValue;
	NSInvocation *am_getBackgroundInvocation = [NSInvocation invocationWithMethodSignature:[NSBezierPath methodSignatureForSelector:@selector(bezierPathWithRect:)]];
	[am_getBackgroundInvocation setTarget:[NSBezierPath class]];
	[am_getBackgroundInvocation setSelector:am_getBackgroundSelector];
	[am_getBackgroundInvocation setArgument:&innerRect atIndex:2];
	[am_getBackgroundInvocation invoke];
	[am_getBackgroundInvocation getReturnValue:&returnValue];
	[self setBackgroundPath:returnValue];
	
	// bezier path for pressed button (with gap for shadow)
	innerRect.size.height--;
	if ([controlView isFlipped]) {
		innerRect.origin.y++;
	}

	[am_getBackgroundInvocation setArgument:&innerRect atIndex:2];
	[am_getBackgroundInvocation invoke];
	[am_getBackgroundInvocation getReturnValue:&returnValue];
	[self setHighlightedBackgroundPath:returnValue];

	// arrow path
	if ([self menu]) {
		float arrowWidth = [NSFont systemFontSizeForControlSize:[self controlSize]]*0.6;
		float arrowHeight = [NSFont systemFontSizeForControlSize:[self controlSize]]*0.5;
		// clip text rect
		am_textRect.size.width -= (radius*0.5)+(arrowWidth/2.0);
		float x = am_lastFrameSize.width-am_backgroundInset-(radius*0.5)-arrowWidth;
		float y = am_backgroundInset+radius-arrowHeight/2.0;

		NSBezierPath *path = [NSBezierPath bezierPath];
		NSPoint point1;
		NSPoint point2;
		NSPoint point3;
		if ([controlView isFlipped]) {
			y += radius*0.05;
			point1 = NSMakePoint(x, y);
			point2 = NSMakePoint(x+arrowWidth, y);
			point3 = NSMakePoint(x+arrowWidth/2.0, y+arrowHeight);
		} else {
			y -= radius*0.05;
			point1 = NSMakePoint(x, y+arrowHeight);
			point2 = NSMakePoint(x+arrowWidth, y+arrowHeight);
			point3 = NSMakePoint(x+arrowWidth/2.0, y);
		}
		[path moveToPoint:point1];
		[path lineToPoint:point3];
		[path lineToPoint:point2];
		[path lineToPoint:point1];
		[path closePath];
		[self setArrowPath:path];
	}
}

- (NSSize)cellSize
{
	return [self intrinsicContentSize];
}

- (NSSize)cellSizeForBounds:(NSRect)aRect
{
	return [self cellSize];
}

- (NSRect)drawingRectForBounds:(NSRect)theRect
{
	NSRect result = [super drawingRectForBounds:theRect];
//	NSLog(@"drawingRectForBounds: %f %f %f %f = %f %f %f %f", theRect.origin.x, theRect.origin.y, theRect.size.width, theRect.size.height, result.origin.x, result.origin.y, result.size.width, result.size.height);
	return result;
}

- (NSSize)intrinsicContentSize
{
	NSSize result = NSZeroSize;
	NSDictionary *textAttributes = [NSDictionary dictionaryWithObjectsAndKeys:[self font], NSFontAttributeName, nil];
	result.width = [[self title] sizeWithAttributes:textAttributes].width;
	switch ([self controlSize]) {
		case NSSmallControlSize: {
			result.height = 19;
			break;
		}
		case NSMiniControlSize: {
			result.height = 16;
			break;
		}
		default: { // NSRegularControlSize
			result.height = 22;
			break;
		}
	}
	CGFloat radius = (result.height/2.0)-am_backgroundInset;
	CGFloat textInset = ceilf([self textInsetForHeight:result.height]);
	result.width += 2.0*(textInset+am_backgroundInset+am_textGap);
	if ([self menu] != nil) {
		float arrowWidth = [NSFont systemFontSizeForControlSize:[self controlSize]]*0.6;
		result.width += (radius*0.5)+arrowWidth;
	}
//	NSLog(@"intrinsic content size width: %f", result.width);
	return result;
}

- (CGFloat)textInsetForHeight:(CGFloat)height
{
	float result;
	float radius = (height/2.0)-am_backgroundInset;
	if ([self amBezelStyle] != NSShadowlessSquareBezelStyle) {
		float h = [[self font] ascender]/2.0;
		//		result = ceilf(radius-sqrt(radius*radius - h*h)+(radius*0.25));
		result = ceilf(radius-sqrt(abs(radius*radius - h*h)));
	} else {
		//		result = ceilf([font ascender]-([font fixed_xHeight]/2.0));
		result = ceilf(am_roundedRectCornerRadius);
	}
	return result;
}

- (NSPoint)menuPositionForFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	NSPoint result = [controlView convertPoint:cellFrame.origin toView:nil];
	result.x += 1.0;
	result.y -= am_backgroundInset+4.0;
	return result;
}


// ======================================================
#pragma mark - Drawing
// ======================================================

- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	if ((am_lastFrameSize.width != cellFrame.size.width) || (am_lastFrameSize.height != cellFrame.size.height)) {
		[self calculateLayoutForFrame:cellFrame inView:controlView];
	}
	[self drawInteriorWithFrame:cellFrame inView:controlView];
	if ([self isBordered]) {
//		if ([self isHighlighted] && ([self highlightsBy] & (NSPushInCellMask | NSChangeGrayCellMask | NSChangeBackgroundCellMask))) {
		if ([self isHighlighted]) {
			[[self highlightedFrameColor] set];
		} else if (am_mouseOver) {
			[[self mouseoverFrameColor] set];
//		} else if (([self state] == NSOnState) && ([self showsStateBy] & (NSChangeGrayCellMask | NSChangeBackgroundCellMask))) {
		} else if ([self state] == NSOnState) {
			[[self highlightedFrameColor] set];
		} else {
			if ([self isEnabled]) {
				[[self frameColor] set];
			} else {
				[[[self frameColor] disabledColor] set];
			}
		}
		// translate to current origin
		NSBezierPath *path = [[[self backgroundPath] copy] autorelease];
		NSAffineTransform *transformation = [NSAffineTransform transform];
		[transformation translateXBy:cellFrame.origin.x+am_backgroundInset yBy:cellFrame.origin.y+am_backgroundInset];
		[path transformUsingAffineTransform:transformation];
		[path setLineWidth:1.0];

		[NSGraphicsContext saveGraphicsState];
		// adjust clipping rectangle
		NSRect halfFrame = cellFrame;
		halfFrame.size.height =  floorf(halfFrame.size.height/2);
		[NSBezierPath clipRect:halfFrame];
		if (![controlView isFlipped]) {
			[[self controlShadow] set];
		}
		[path stroke];
		[NSGraphicsContext restoreGraphicsState];
		[NSGraphicsContext saveGraphicsState];
		halfFrame.origin.y = cellFrame.origin.y+halfFrame.size.height;
		[NSBezierPath clipRect:halfFrame];
		if ([controlView isFlipped]) {
			[[self controlShadow] set];
		}
		[path stroke];
		[NSGraphicsContext restoreGraphicsState];
	}
}

- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
//	NSLog(@"highlightsBy = %i", [self highlightsBy]);
//	NSLog(@"showsStateBy = %i", [self showsStateBy]);
	NSBezierPath *path;
	NSColor *textColor;
	NSColor *arrowColor;
	NSShadow *textShadow;
	NSColor *shadingStartColor= nil;
	NSColor *shadingEndColor = nil;
	BOOL drawShading = NO;
	NSAffineTransform *transformation = [NSAffineTransform transform];
	[transformation translateXBy:cellFrame.origin.x+am_backgroundInset yBy:cellFrame.origin.y+am_backgroundInset];
	if ([self isHighlighted] && ([self highlightsBy] & (NSPushInCellMask | NSChangeGrayCellMask | NSChangeBackgroundCellMask))) {
		[NSGraphicsContext saveGraphicsState];
		if ([self highlightedShadingColor] != nil) {
			drawShading = YES;
			shadingStartColor = [self highlightedShadingColor];
			shadingEndColor = [self highlightedControlColor];
		}
		[[self highlightedControlColor] set];
		[[self highlightedControlShadow] set];
		path = [[[self highlightedBackgroundPath] copy] autorelease];
		textColor = [self highlightedTextColor];
		arrowColor = [self highlightedArrowColor];
		textShadow = [self highlightedTextShadow];
	} else if (am_mouseOver) {
		if ([self mouseoverShadingColor] != nil) {
			drawShading = YES;
			shadingStartColor = [self mouseoverShadingColor];
			shadingEndColor = [self mouseoverControlColor];
		}
		[[self mouseoverControlColor] set];
		path = [[[self backgroundPath] copy] autorelease];
		textColor = [self mouseoverTextColor];
		arrowColor = [self mouseoverArrowColor];
		textShadow = [self mouseoverTextShadow];
	} else if (([self state] == NSOnState) && ([self showsStateBy] & (NSChangeGrayCellMask | NSChangeBackgroundCellMask))) {
		if ([self highlightedShadingColor] != nil) {
			drawShading = YES;
			shadingStartColor = [self highlightedShadingColor];
			shadingEndColor = [self highlightedControlColor];
		}
		[[self highlightedControlColor] set];
		path = [[[self backgroundPath] copy] autorelease];
		textColor = [self highlightedTextColor];
		arrowColor = [self highlightedArrowColor];
		textShadow = [self highlightedTextShadow];
	} else {
		if ([self shadingColor] != nil) {
			drawShading = YES;
			if ([self isEnabled]) {
				shadingStartColor = [self shadingColor];
				shadingEndColor = [self controlColor];
			} else {
				shadingStartColor = [[self shadingColor] disabledColor];
				shadingEndColor = [[self controlColor] disabledColor];
			}
		}
		[[self controlColor] set];
		path = [[[self backgroundPath] copy] autorelease];
		if ([self isEnabled]) {
			textColor = [self textColor];
			arrowColor = [self arrowColor];
		} else {
			textColor = [[self textColor] disabledColor];
			arrowColor = [[self arrowColor] disabledColor];
		}
		textShadow = [self textShadow];
	}
	
	[path transformUsingAffineTransform:transformation];
	[path setLineWidth:0.0];
	[path setFlatness:am_bezierPathFlatness];
	if (drawShading) {
		switch (self.shadingMode) { 
			case AMRollOverButtonShadingLinear: {
				if ([controlView isFlipped]) {
					[path linearGradientFillWithStartColor:shadingEndColor endColor:shadingStartColor];
				} else {
					[path linearGradientFillWithStartColor:shadingStartColor endColor:shadingEndColor];
				}
				break;
			}
			case AMRollOverButtonShadingBilinear: {
				[path bilinearGradientFillWithOuterColor:shadingStartColor innerColor:shadingEndColor];
				break;
			}
			default: // am_rollOverButtonShadingNone
			{
				[path fill];
			}
		}
	} else {
		[path fill];
	}

	if ([self isHighlighted]) {
		[NSGraphicsContext restoreGraphicsState];
	}
	[NSGraphicsContext saveGraphicsState];
	[textShadow set];
	// menu arrow
	if (([self menu] && self.alwaysShowMenuIndicator) || ([self menu] && am_mouseOver)) {
		// draw menu arrow
		if ([self isEnabled]) {
			[arrowColor set];
		} else {
			[[arrowColor disabledColor] set];
		}
		[[self arrowPath] fill];
	}
	// draw button title
	NSDictionary *stringAttributes;
	NSFont *font;
	NSMutableParagraphStyle *parapraphStyle = [[[NSMutableParagraphStyle alloc] init] autorelease];
	[parapraphStyle setAlignment:[self alignment]];
	[parapraphStyle setLineBreakMode:NSLineBreakByTruncatingTail];
	font = [self font];
	stringAttributes = [NSDictionary dictionaryWithObjectsAndKeys:font, NSFontAttributeName, textColor, NSForegroundColorAttributeName, parapraphStyle, NSParagraphStyleAttributeName, nil];
	NSString *currentTitle = nil;
	if (([self state] == NSOnState) && ([self showsStateBy] & NSContentsCellMask)) {
		currentTitle = [self alternateTitle];
	}
	if ([self isHighlighted] && ([self highlightsBy] & NSContentsCellMask)) {
		currentTitle = [self alternateTitle];
	}
	if ((currentTitle == nil) || ([currentTitle length] == 0)){
		currentTitle = [self title];
	}
	[currentTitle drawInRect:am_textRect withAttributes:stringAttributes];
	[NSGraphicsContext restoreGraphicsState];
}


// ======================================================
#pragma mark - Event handling
// ======================================================

- (void)showMenuForEvent:(NSEvent *)theEvent controlView:(NSView *)controlView cellFrame:(NSRect)cellFrame
{
	NSPoint menuPosition = [self menuPositionForFrame:cellFrame inView:controlView];
	// create event for pop up menu with adjusted mouse position
	NSEvent *menuEvent = [NSEvent mouseEventWithType:[theEvent type] location:menuPosition modifierFlags:[theEvent modifierFlags] timestamp:[theEvent timestamp] windowNumber:[theEvent windowNumber] context:[theEvent context] eventNumber:[theEvent eventNumber] clickCount:[theEvent clickCount] pressure:[theEvent pressure]];
	/* tried to determine the menus size - no way  :-(
		void *carbonMenu = [[self menu] _menuImpl];
	NSLog(@"[self menu]: %@", [self menu]);
	NSLog(@"_menuImpl: %@", carbonMenu);
	NSLog(@"menu: %@", [carbonMenu menu]);
	int menuHeight = GetMenuHeight(carbonMenu);
	NSLog(@"height: %i", menuHeight);
	NSRect outRect;
	OSStatus status = HIViewGetBounds([[self menu] _menuImpl], &outRect);
	NSLog(@"status: %li", status);
	NSLog(@"width: %f height: %f", outRect.size.width, outRect.size.height);
	*/
	[NSMenu popUpContextMenu:[self menu] withEvent:menuEvent forView:controlView];
}

- (BOOL)trackMouse:(NSEvent *)theEvent inRect:(NSRect)cellFrame ofView:(NSView *)controlView untilMouseUp:(BOOL)untilMouseUp
{
	
	BOOL result = NO;
	NSDate *endDate;
	NSPoint currentPoint = [theEvent locationInWindow];
	BOOL done = NO;
	if ([self menu]) {
		// check if mouse is over menu arrow
		NSPoint localPoint = [controlView convertPoint:currentPoint fromView:nil];
		if (localPoint.x >= (am_textRect.origin.x+am_textRect.size.width)) {
			done = YES;
			result = YES;
			[self showMenuForEvent:theEvent controlView:controlView cellFrame:cellFrame];
		}
	}
	BOOL trackContinously = [self startTrackingAt:currentPoint inView:controlView];
	// catch next mouse-dragged or mouse-up event until timeout
	BOOL mouseIsUp = NO;
	NSEvent *event;
	while (!done) { // loop ...
		NSPoint lastPoint = currentPoint;
		if ([self menu]) { // timeout for menu
			endDate = [NSDate dateWithTimeIntervalSinceNow:self.popUpMenuDelay];
		} else { // no timeout
			endDate = [NSDate distantFuture];
		}
		event = [NSApp nextEventMatchingMask:(NSLeftMouseUpMask|NSLeftMouseDraggedMask) untilDate:endDate inMode:NSEventTrackingRunLoopMode dequeue:YES];
		if (event) { // mouse event
			currentPoint = [event locationInWindow];
			if (trackContinously) { // send continueTracking.../stopTracking...
				if (![self continueTracking:lastPoint at:currentPoint inView:controlView]) {
					done = YES;
					[self stopTracking:lastPoint at:currentPoint inView:controlView mouseIsUp:mouseIsUp];
				}
				if ([self isContinuous]) {
					[NSApp sendAction:[self action] to:[self target] from:controlView];
				}
			}
			mouseIsUp = ([event type] == NSLeftMouseUp);
			done = done || mouseIsUp;
			if (untilMouseUp) {
				result = mouseIsUp;
			} else {
				// check, if the mouse left our cell rect
				result = NSPointInRect([controlView convertPoint:currentPoint fromView:nil], cellFrame);
				if (!result) {
					done = YES;
					[self setMouseOver:NO];
				} else {
					[self setMouseOver:[(AMRollOverButton *)controlView tracksMouseInside]];
				}
			}
			if (done && result && ![self isContinuous]) {
				[NSApp sendAction:[self action] to:[self target] from:controlView];
				if ([(AMRollOverButton *)self.controlView buttonType] == NSOnOffButton) {
					[self setNextState];
				}
			}
		} else { // show menu
			done = YES;
			result = YES;
			[self showMenuForEvent:theEvent controlView:controlView cellFrame:cellFrame];
		}
	} // while (!done)
	return result;
}

@end
